__version__ = "0.8.1"

from .base import Wrapped
from .numpy import (
    from_ndarray as np_from_ndarray,
    from_generator as np_from_generator,
    open_existing as np_open_existing,
    extend_dir as np_extend_dir,
    extend as np_extend,
    append as np_append
)

from .generic import open_existing
from .ragged import RaggedMmap
from .string import StringsMmap



import struct
from copy import copy
from dataclasses import dataclass
from pathlib import Path
from typing import Union, Sequence, List

from .parallel import ParallelBatchCollector


def _bytes_to_int(inp: bytes, fmt: str = "<i") -> int:
    """
    Utility function for converting bytes into int.

    :param inp: the ``bytes`` input
    :param fmt: the format of the integer. Check https://docs.python.org/3/library/struct.html#format-characters
    :return: The integer
    """
    return struct.unpack(fmt, inp)[0]


def _int_to_bytes(inp: int, fmt: str = "<i") -> bytes:
    """
    Stores an integer as bytes

    :param inp: The integer input
    :param fmt: the format to use when storing. Check https://docs.python.org/3/library/struct.html#format-characters
    :return:
    """
    return struct.pack(fmt, inp)


def _bytes_to_str(inp: bytes, encoding: str = "utf-8") -> str:
    """
    Reads bytes into a string.

    :param inp: The bytes array
    :param encoding: The encoding to use.
    :return: The string
    """
    return inp.decode(encoding)


def _str_to_bytes(inp: str, encoding: str = "utf-8") -> bytes:
    """
    Dumps string as bytes

    :param inp: The string input
    :param encoding: The encoding to use
    :return: the byte array
    """
    return inp.encode(encoding)


def _int_to_file(inp: int, file: Union[str, Path], *args, **kwargs) -> None:
    """
    Utility function, which writes an integer as bytes in a given file.

    :param inp: The integer to be stored
    :param file: The file in which the integer should be stored.
    :param args: additional positional args to be passed to ``int_to_bytes``
    :param kwargs: additional keyword args to be passed to ``int_to_bytes``
    :return:
    """
    with open(file, "wb") as out_file:
        out_file.write(_int_to_bytes(inp, *args, **kwargs))


def _file_to_int(file: Union[str, Path], *args, **kwargs) -> int:
    """
    Reads an integer from a file

    :param file: The file to read
    :param args: additional positional args to be passed to ``bytes_to_int``
    :param kwargs: additional keyword args to be passed to ``bytes_to_int``
    :return: the integer
    """
    with open(file, "rb") as in_file:
        return _bytes_to_int(in_file.read(), *args, **kwargs)


def _str_to_file(inp: str, file: Union[str, Path], *args, **kwargs) -> None:
    """
    Stores a string to a file

    :param inp: The string to be written
    :param file: The output file
    :param args: additional positional args to be passed to ``str_to_bytes``
    :param kwargs: additional keyword args to be passed to ``str_to_bytes``
    :return:
    """
    with open(file, "wb") as out_file:
        out_file.write(_str_to_bytes(inp, *args, **kwargs))


def _file_to_str(file: Union[str, Path], *args, **kwargs) -> str:
    """
    Read string from file.

    :param file: The file to read from
    :param args: additional positional args to be passed to ``bytes_to_str``
    :param kwargs: additional keyword args to be passed to ``bytes_to_str``
    :return: The string
    """
    with open(file, "rb") as in_file:
        return _bytes_to_str(in_file.read(), *args, **kwargs)


def _shape_to_bytes(shape: Sequence[int], fmt: str = "<Q") -> bytes:
    """
    Returns the shape of numpy array as a byte array.

    :param shape: The shape of the numpy array
    :param fmt: The format of the bytes
    :return: The bytes array
    """
    res = bytearray()
    for axis_len in shape:
        res.extend(_int_to_bytes(axis_len, fmt=fmt))
    return bytes(res)


def _bytes_to_shape(inp: bytes, step=8, fmt: str = "<Q") -> Sequence[int]:
    """
    Reads shape tuple from bytes.

    :param inp: the bytes array
    :param step: Number of bytes to read for an int
    :param fmt: The formatting used for the integer.
    :return: A tuple of integers, representing the shape.
    """
    res = []
    for start in range(0, len(inp) - 1, step):
        end = start + step
        res.append(_bytes_to_int(inp[start:end], fmt=fmt))
    return tuple(res)


def _shape_to_file(shape: Sequence[int], file: Union[str, Path]) -> None:
    """
    Write a shape sequence to a file

    :param shape: The shape to be dumped.
    :param file: The output file
    :return:
    """
    with open(file, "wb") as out_file:
        out_file.write(_shape_to_bytes(shape))


def _file_to_shape(file: Union[str, Path]) -> Sequence[int]:
    """
    Read shape from file

    :param file: The input file
    :return: the shape as a tuple
    """
    with open(file, "rb") as out_file:
        return _bytes_to_shape(out_file.read())


@dataclass
class BytesSlices:
    buffer: bytes
    starts: List[int]
    ends: List[int]


def _sequence_of_strings_to_bytes(strings: Sequence[str], verbose=False) -> BytesSlices:
    """
    Converts a sequence of strings to a bytes array

    :param strings: A sequence of strings.
    :param verbose: whether a tqdm progressbar should be shown.
    :return: the bytes slices
    """
    buffer = bytearray()
    starts = []
    ends = []
    if verbose:
        from tqdm import tqdm

        strings = tqdm(strings)
    for string in strings:
        arr = _str_to_bytes(string)
        starts.append(len(buffer))
        ends.append(len(buffer) + len(arr))
        buffer.extend(arr)
    return BytesSlices(bytes(buffer), starts, ends)


def from_generator_base(out_dir, sample_generator, batch_size, batch_ctor, extend_fn=None, **kwargs):
    """
    Creates an output from a generator, flushing every batch to disk.

    :param out_dir: The output directory.
    :param sample_generator: The generator of samples.
    :param batch_size: The batch size, which controls how often the output should be written to disk.
    :param batch_ctor: The constructor used to initialize the output.
    :param extend_fn: Functon to call when doing .extend. By default, this will call memmap.extend(samples)
    :param kwargs: Additional keyword arguments to be passed when initializing the output.
    :return:
    """
    out_dir = Path(out_dir)
    out_dir.mkdir(exist_ok=True)
    samples = []
    memmap = None
    if kwargs.pop("verbose", False):
        from tqdm.auto import tqdm

        sample_generator = tqdm(sample_generator)
    for sample in sample_generator:
        samples.append(sample)
        if len(samples) % batch_size != 0:
            continue
        if memmap is None:
            memmap = batch_ctor(out_dir, samples, **kwargs)
        else:
            if extend_fn is not None:
                extend_fn(memmap, samples)
            else:
                memmap.extend(samples)
        samples = []
    if len(samples) > 0:
        if memmap is None:
            memmap = batch_ctor(out_dir, samples, **kwargs)
        else:
            if extend_fn is not None:
                extend_fn(memmap, samples)
            else:
                memmap.extend(samples)
    return memmap


def from_indexable_base(out_dir, indexable, batch_size, batch_ctor, extend_fn=None, n_jobs=None, verbose=False, **kwargs):
    """
    Creates an output from an indexable object, flushing every batch to disk. Can be done in parallel.
    indexable[i] (or indexable(i)) will become memmap[i].

    :param out_dir: The output directory.
    :param indexable: An object that supports __getitem__ or a function that takes one integer argument.
    :param batch_size: The batch size, which controls how often the output should be written to disk.
    :param batch_ctor: The constructor used to initialize the output.
    :param extend_fn: Functon to call when doing .extend. By default, this will call memmap.extend(samples).
    :param n_jobs: number of jobs to iterate through indexable with. Default=None corresponds to no parallelization.
    :param verbose: whether to print progress meter.
    :param kwargs: Additional keyword arguments to be passed when initializing the output.
    :return:
    """
    out_dir = Path(out_dir)
    out_dir.mkdir(exist_ok=True)
    memmap = None

    batch_collector = ParallelBatchCollector(indexable, batch_size, n_jobs, verbose)

    for samples in batch_collector.batches():
        if memmap is None:
            memmap = batch_ctor(out_dir, samples, **kwargs)
        else:
            if extend_fn is not None:
                extend_fn(memmap, samples)
            else:
                memmap.extend(samples)

    return memmap


class Wrapped:
    def __init__(self, data, wrapper_fn, copy_before_wrapper_fn=True):
        self.data = data
        self.wrapper_fn = wrapper_fn
        self.copy_before_wrapper_fn = copy_before_wrapper_fn

    def __getitem__(self, item):
        sample = self.data[item]
        if self.copy_before_wrapper_fn:
            sample = copy(sample)
        return self.wrapper_fn(sample)

    def __len__(self):
        return len(self.data)



from enum import Enum
from functools import partial
from tqdm.auto import tqdm

try:
    from joblib import Parallel, delayed
except ImportError:
    Parallel, delayed = None, None
    HAS_JOBLIB = False
else:
    HAS_JOBLIB = True


class _Exhausted(Enum):
    exhausted = 'EXHAUSTED'


EXHAUSTED = _Exhausted.exhausted


class ParallelBatchCollector:
    _parallel: Parallel = None

    def __init__(self, indexable, batch_size, n_jobs=None, verbose=False, **kwargs):
        self.indexable, self._obj_length, self._num_batches = self.verify(indexable, batch_size)
        self.batch_size = batch_size

        self._pbar = self._init_pbar(verbose)
        self._parallel = self.begin(n_jobs, **kwargs)
        self._batch_num = 0
        self._exhausted = False

    @staticmethod
    def verify(indexable, batch_size):
        try:
            _ = indexable.__getitem__
        except AttributeError:
            if callable(indexable):
                indexable = _IndexableWrap(indexable)
            else:
                msg = 'indexable must implement __getitem__ or be callable and take one integer argument.'
                raise TypeError(msg)

        try:
            length = len(indexable)
        except TypeError:
            length = None
            num_batches = None
        else:
            num_batches = length // batch_size + (length % batch_size != 0)

        return indexable, length, num_batches

    @staticmethod
    def begin(n_jobs: int, **kwargs):
        if n_jobs in (None, 1):
            return
        elif not HAS_JOBLIB:
            msg = 'joblib is not installed. Install joblib or run with n_jobs=None to ignore parallelization.'
            raise ImportError(msg)

        _parallel = Parallel(n_jobs=n_jobs, **kwargs)
        _parallel.__enter__()
        return _parallel

    def batches(self):
        while not self.exhausted():
            yield self.collect_batch()

    def collect_batch(self):
        if self._parallel is None:
            batch = self._collect_no_parallel_batch()
        else:
            batch = self._collect_parallel_batch()

        self._update_pbar(batch)
        return batch

    def _init_pbar(self, verbose):
        if not verbose:
            return None
        return tqdm(total=self._obj_length)

    def _update_pbar(self, batch):
        if self._pbar is not None:
            self._pbar.update(len(batch))

    def _collect_no_parallel_batch(self):
        results = [_get_from_indexable(self.indexable, j) for j in self._rng()]

        if self.exhausted(results):
            results = [r for r in results if r is not EXHAUSTED]

        return results

    def _collect_parallel_batch(self):
        func = delayed(partial(_get_from_indexable, self.indexable))

        results = self._parallel(func(j) for j in self._rng())

        if self.exhausted(results):
            results = [r for r in results if r is not EXHAUSTED]
            self._parallel.__exit__(None, None, None)

        return results

    def exhausted(self, results=()):
        self._exhausted = (
                self._exhausted or
                any(r is EXHAUSTED for r in results) or
                self.completed_batches()
        )

        return self._exhausted

    def completed_batches(self):
        return self._num_batches is not None and self._batch_num == self._num_batches

    def _rng(self):
        start = self.batch_size * self._batch_num
        stop = self.batch_size * (1 + self._batch_num)

        self._batch_num += 1

        return range(start, stop)


class _IndexableWrap:
    def __init__(self, func):
        self._func = func

    def __getitem__(self, item):
        return self._func(item)

    @property
    def wrapped(self):
        return self._func


class _IndexableLengthWrap(_IndexableWrap):
    def __init__(self, func, length):
        super().__init__(func)
        self.length = length

    def __len__(self):
        return self.length


def make_indexable(func, length=None):
    if length is not None:
        return _IndexableLengthWrap(func, length)
    return _IndexableWrap(func)


def _get_from_indexable(indexable, item,):
    try:
        return indexable[item]
    except (IndexError, KeyError):
        return EXHAUSTED



import os
import stat

import numpy as np
from mmap_ninja import generic, numpy as np_ninja


def test_numpy(tmp_path):
    arr = np.arange(10)
    memmap = np_ninja.from_ndarray(tmp_path / "numpy_mmap", arr)

    for i, el in enumerate(arr):
        assert el == memmap[i]


def test_numpy_from_empty(tmp_path):
    arr = np.arange(10)
    memmap = np_ninja._empty(tmp_path / "numpy_mmap", dtype=np.int64, shape=(10,), order="C")
    memmap[:] = arr


def test_numpy_extend(tmp_path):
    arr = np.arange(3)
    memmap = np_ninja.from_ndarray(tmp_path / "growable", arr)
    np_ninja.extend_dir(tmp_path / "growable", np.arange(11, 13))
    memmap = np_ninja.open_existing(tmp_path / "growable")
    assert memmap[0] == 0
    assert memmap[1] == 1
    assert memmap[2] == 2
    assert memmap[3] == 11
    assert memmap[4] == 12

    memmap2 = generic.open_existing(tmp_path / "growable")
    assert np.all(memmap == memmap2)


def test_numpy_append(tmp_path):
    arr = np.arange(3)
    memmap = np_ninja.from_ndarray(tmp_path / "growable", arr)
    np_ninja.append(memmap, np.asarray(4))
    memmap = np_ninja.open_existing(tmp_path / "growable")
    assert memmap[-1] == 4


def test_numpy_extend_alternative_api(tmp_path):
    arr = np.arange(3)
    memmap = np_ninja.from_ndarray(tmp_path / "growable", arr)
    np_ninja.extend(memmap, np.arange(11, 13))
    memmap = np_ninja.open_existing(tmp_path / "growable")
    assert memmap[0] == 0
    assert memmap[1] == 1
    assert memmap[2] == 2
    assert memmap[3] == 11
    assert memmap[4] == 12

    memmap2 = generic.open_existing(tmp_path / "growable", wrapper_fn=lambda x: x / 20.0)
    assert np.allclose(memmap / 20.0, memmap2)


def simple_gen():
    for i in range(30):
        yield i


def test_numpy_from_generator(tmp_path):
    memmap = np_ninja.from_generator(
        out_dir=tmp_path / "generator", sample_generator=simple_gen(), batch_size=4, verbose=True
    )
    memmap = np_ninja.open_existing(tmp_path / "generator")
    for i in range(30):
        assert i == memmap[i]


def test_read_only(tmp_path):
    out_path = tmp_path / "read_only_numpy_memmap"
    arr = np.arange(10)
    np_ninja.from_ndarray(out_path, arr)
    # Make all memmap files read-only for the user
    for p in out_path.glob(r"**/*"):
        if not p.is_file():
            continue
        os.chmod(p, stat.S_IRUSR)
    # Open in read-only mode
    memmap = np_ninja.open_existing(out_path, mode="r")
    for i, el in enumerate(arr):
        assert el == memmap[i]
